from datetime import timedelta

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model, update_session_auth_hash
from django.contrib.auth.forms import PasswordChangeForm, PasswordResetForm, SetPasswordForm
from django.contrib.auth.password_validation import validate_password
from django.contrib.auth.tokens import default_token_generator
from django.contrib.auth.views import PasswordResetConfirmView
from django.contrib.sites.shortcuts import get_current_site
from django.core import signing
from django.shortcuts import reverse
from django.template.loader import render_to_string
from django.utils.encoding import force_bytes, force_text
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import ugettext_lazy as _


from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from .auth import generate_verify_email_token
from .models import ImageAsset, Article, ArticleCollection, ArticleOrder, ArticleCollectionOrder, RichTextAsset, TextAsset
from .models import User

user_model = get_user_model()


class UserSerializer(serializers.ModelSerializer):

    is_admin = serializers.BooleanField(source='is_superuser', read_only=True)

    def update(self, instance, validated_data):

        email_changed = False
        
        if not self.partial and validated_data.get('email') is not None:

            email_changed = instance.email != validated_data['email']

        if email_changed:
            instance.is_email_verified = False

        instance = super().update(instance, validated_data)

        if email_changed:
            try:
                instance.send_verification_email()
            except:
                # don't want email send exceptions to block updates
                pass

        return instance

    class Meta:
        model = user_model
        fields = ('id', 'email', 'first_name', 'last_name', 'profile_pic',
                  'phone_number', 'is_authenticated', 'is_admin', 'is_staff',
                  'is_email_verified')
        read_only_fields = ['is_authenticated',
                            'is_admin', 'is_staff', 'is_email_verified']


class ImageAssetSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImageAsset

        fields = '__all__'


class ArticleSerializer(serializers.ModelSerializer):
    owner = UserSerializer(read_only=True)

    class Meta:
        model = Article
        depth = 1
        fields = '__all__'


class ArticleOrderSerializer(serializers.ModelSerializer):
    class Meta:
        model = ArticleOrder

        fields = ['article', 'order']


class ArticleCollectionOrderSerializer(serializers.ModelSerializer):
    class Meta:
        model = ArticleCollectionOrder

        fields = ['article_collection', 'order']


class ArticleCollectionSerializer(serializers.ModelSerializer):

    article_order = ArticleOrderSerializer(
        source='articleorder_set', many=True)
    article_collection_order = ArticleCollectionOrderSerializer(
        source='container_articles', many=True)
    owner = UserSerializer(read_only=True)

    class Meta:
        model = ArticleCollection
        depth = 1
        fields = '__all__'


class RichTextAssetSerializer(serializers.ModelSerializer):
    class Meta:
        model = RichTextAsset

        fields = '__all__'


class TextAssetSerializer(serializers.ModelSerializer):
    class Meta:
        model = TextAsset

        fields = '__all__'


class ImageAssetSerializerB(serializers.ModelSerializer):
    class Meta:
        model = ImageAsset

        fields = ('assetid', 'id', 'image', 'published')


class RichTextAssetSerializerB(serializers.ModelSerializer):
    class Meta:
        model = RichTextAsset

        fields = ('assetid', 'id', 'text', 'published')


class TextAssetSerializerB(serializers.ModelSerializer):
    class Meta:
        model = TextAsset

        fields = ('assetid', 'id', 'text', 'published')


class ImageAssetSerializerC(serializers.ModelSerializer):
    class Meta:
        model = ImageAsset

        fields = ('assetid',)


class RichTextAssetSerializerC(serializers.ModelSerializer):
    class Meta:
        model = RichTextAsset

        fields = ('assetid',)


class TextAssetSerializerC(serializers.ModelSerializer):
    class Meta:
        model = TextAsset

        fields = ('assetid',)


class ArticleSerializerB(serializers.ModelSerializer):

    title = TextAssetSerializerC(read_only=True)
    text = RichTextAssetSerializerC(read_only=True)
    image = ImageAssetSerializerC(read_only=True)

    class Meta:
        model = Article
        depth = 1
        fields = ('assetid', 'id', 'title', 'text', 'image', 'published')


class ArticleCollectionSerializerB(serializers.ModelSerializer):

    class Meta:
        model = ArticleCollection
        fields = ('assetid',)


class SignUpSerializer(serializers.ModelSerializer):

    password_confirm = serializers.CharField(write_only=True)
    first_name = serializers.CharField(required=True)

    class Meta:
        model = user_model
        fields = ['email', 'password', 'first_name', 'phone_number',
                  'last_name', 'password_confirm']
        extra_kwargs = {
            'password': {'write_only': True},
        }

    def validate(self, attrs):
        # check password match
        if attrs.get('password') != attrs.get('password_confirm'):
            raise ValidationError(_('Password mismatch'))

        attrs.pop('password_confirm')
        # do password validation check here as we want mismatch
        # to fire before password not allowed

        user = user_model(**attrs)
        try:
            validate_password(attrs.get('password'), user)

        except ValidationError as validation_error:
            raise ValidationError({'password': validation_error.messages})

        return attrs

    def create(self, validated_data):
        user = user_model.objects.create_user(**validated_data)
        return user

    def save(self, **kwargs):

        super().save(**kwargs)

        # self.instance.send_sign_up_email()


class LoginSerializer(serializers.Serializer):

    email = serializers.CharField(required=True)
    password = serializers.CharField(required=True)

    def validate(self, data):

        user = authenticate(self.context['request'], **data)

        # add some specific checks here like user.is_active
        if user:
            if not user.is_active:
                err_msg = _('User account is disabled')
                raise ValidationError(err_msg)

            data['user'] = user
            return data

        raise ValidationError(_('Unable to login with provided credentials'))


class PasswordChangeSerializer(serializers.Serializer):

    old_password = serializers.CharField(required=True)
    new_password1 = serializers.CharField(required=True)
    new_password2 = serializers.CharField(required=True)

    def __init__(self, *args, **kwargs):

        # need to pass user information to the PasswordChangeForm
        super(PasswordChangeSerializer, self).__init__(*args, **kwargs)
        self.request = self.context.get('request')

        self.user = getattr(self.request, 'user', None)

    def validate_old_password(self, value):

        if not self.user.check_password(value):
            raise serializers.ValidationError(
                _("Current password provided was incorrect."))

        return value

    def validate(self, data):

        self.password_change_form = PasswordChangeForm(self.user, data=data)

        if not self.password_change_form.is_valid():
            raise ValidationError(self.password_change_form.errors)

        return data

    def save(self):

        self.password_change_form.save()
        update_session_auth_hash(self.request, self.user)


class PasswordResetRequestSerializer(serializers.Serializer):

    email = serializers.EmailField()

    def validate_email(self, value):

        reset_form = PasswordResetForm(data=self.initial_data)

        if not reset_form.is_valid():
            raise serializers.ValidationError(reset_form.errors)

        return value

    def get_user(self):

        # email is unique but call .filter().first()
        # to prevent throwing here

        active_user = user_model.objects.filter(
            email__iexact=self.validated_data['email'],
            is_active=True
        ).first()

        if active_user and active_user.has_usable_password:
            return active_user

        return None

    def save(self):

        user = self.get_user()
        # TODO to handle no user found throw ValidationError ?

        if user:
            user.send_password_reset_email()


class PasswordResetConfirmSerializer(serializers.Serializer):

    new_password1 = serializers.CharField(max_length=256)
    new_password2 = serializers.CharField(max_length=256)
    uidb64 = serializers.CharField()
    token = serializers.CharField()

    def validate(self, data):
        self._errors = {}
        # first get user from uid
        try:
            uid = force_text(urlsafe_base64_decode(data['uidb64']))
            self.user = user_model.objects.get(pk=uid)
        except (TypeError, ValueError, OverflowError, user_model.DoesNotExist):
            raise ValidationError({'uidb64': ['Invalid value']})

        self.set_password_form = SetPasswordForm(
            user=self.user, data=data
        )

        if not self.set_password_form.is_valid():
            raise ValidationError(self.set_password_form.errors)
        if not default_token_generator.check_token(self.user, data['token']):
            raise ValidationError(
                {'token': ['Expiried or is invalid']})

        return data

    def save(self):
        return self.set_password_form.save()


# class SendEmailVerificationSerializer(serializers.Serializer):

#     user_id = serializers.IntegerField()
#     email = serializers.EmailField()

#     def save(self):

#         token = generate_verify_email_token(
#             self.validated_data['user_id'],
#             self.validated_data['email']
#         )

#         # TODO red - this is current hard coded, maybe make it dynamic
#         template_params = {
#             'user': self.context.get('user'),
#             'url': f"{settings.SITE_URL}/verify?token={token}"
#         }

#         msg_plain = render_to_string('verify_email.txt', template_params)
#         msg_html = render_to_string('verify_email.html', template_params)

#         filtered_send_mail(
#             'Ripple Energy - Account Verification Email',
#             msg_plain,
#             'support@example.com',
#             [self.validated_data['email']],
#             html_message=msg_html,
#             fail_silently=False,
#         )


class ConfirmEmailSerializer(serializers.Serializer):

    token = serializers.CharField(max_length=256)

    def __init__(self, *args, **kwargs):
        super(ConfirmEmailSerializer, self).__init__(*args, **kwargs)

    def validate_token(self, value):

        invalid_token_message = ('Invalid Token: Please ensure you have'
                                 ' used the link'
                                 ' provided in the verification email.')

        try:
            unsigned = signing.loads(value, max_age=timedelta(days=1))

        except signing.SignatureExpired:
            raise ValidationError(_('Verification Token Expired'))
        except signing.BadSignature:
            # token tampered with
            raise ValidationError(
                _(invalid_token_message))

        if not unsigned.get('action') == 'verify_email':
            raise ValidationError(_(invalid_token_message))

        self.user = user_model.objects.get(id=unsigned['user_id'])

        if not self.user.email == unsigned['email']:
            raise ValidationError(
                _(('Email address associated with this account does not match'
                   ' with confirmation token'))
            )

        return value

    def validate(self, attrs):

        if self.user.is_email_verified:
            raise ValidationError(
                _("Email has already been verified for this account"))

        return attrs

    def save(self):

        user = self.user
        user.is_email_verified = True
        user.save()
        return user



import time

from django.core import signing
from django.db import transaction


def generate_token(token_params):

    return signing.dumps(token_params)


def generate_verify_email_token(user_id, email):
    return generate_token({
        'user_id': user_id,
        'email': email,
        'action': 'verify_email',
    })



from django.core import mail
from django.template.loader import render_to_string

from rest_framework import status


class TestEmailMixin:

    def assert_email_has_been_sent(
        self,
        # to,
        # subject,
        template_path_txt,
        template_path_html,
        template_context
    ):
        expected_text = render_to_string(
            template_path_txt, template_context)

        expected_html = render_to_string(
            template_path_html, template_context)

        sent_mail = mail.outbox[0]
        self.assertIsNotNone(sent_mail)

        # self.assertEqual(sent_mail.to, to)
        # self.assertEqual(sent_mail.subject, subject)
        self.assertEqual(sent_mail.body, expected_text)

        sent_html = None
        for alternative in sent_mail.alternatives:
            # (<h1>Some email html</h1>, 'text/html')
            if alternative[1] == 'text/html':
                sent_html = alternative[0]
                break
        self.assertIsNotNone(sent_html)
        self.assertEqual(sent_html, expected_html)


class GenericAPITestCase:

    expected_status_code = status.HTTP_200_OK
    expected_return_json = {}
    method = 'get'

    def setUp(self):
        self.response = getattr(self.client, self.method)(
            self.end_point,
            self.payload
        )

    @property
    def end_point(self):
        raise NotImplementedError()

    @property
    def payload(self):
        return None

    def test_expected_response(self):
        self.assertEqual(self.response.status, self.expected_status_code)
        self.assertDictEqual(self.response.json(), self.expected_return_json)


class TestAuthenticationMixin:

    use_authenticated_client = True

    @classmethod
    def setUpTestData(cls):
        # print(cls.client)
        cls.user = UserFactory()
        cls.not_registered_user = factory.build(
            dict, FACTORY_CLASS=UserFactory)

        super().setUpTestData()

    def setUp(self):

        if self.use_authenticated_client:
            self.client.force_login(self.user)

        super().setUp()




from django import forms
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.forms import ReadOnlyPasswordHashField
from django.utils.translation import ugettext_lazy as _

from django.contrib.auth import password_validation
import unicodedata

from .models import User
from .models import ImageAsset, Article, ArticleCollection, RichTextAsset, TextAsset, ArticleOrder, ArticleCollectionOrder


class UsernameField(forms.CharField):
    def to_python(self, value):
        return unicodedata.normalize('NFKC', super().to_python(value))


class UserCreationForm2(forms.ModelForm):
    """
    A form that creates a user, with no privileges, from the given username and
    password.
    """
    error_messages = {
        'password_mismatch': _("The two password fields didn't match."),
    }
    password1 = forms.CharField(
        label=_("Password"),
        strip=False,
        widget=forms.PasswordInput,
        help_text=password_validation.password_validators_help_text_html(),
    )
    password2 = forms.CharField(
        label=_("Password confirmation"),
        widget=forms.PasswordInput,
        strip=False,
        help_text=_("Enter the same password as before, for verification."),
    )

    class Meta:
        model = User
        fields = ('email',)
        field_classes = {'email': UsernameField}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._meta.model.USERNAME_FIELD in self.fields:
            self.fields[self._meta.model.USERNAME_FIELD].widget.attrs.update({'autofocus': True})

    def clean_password2(self):
        password1 = self.cleaned_data.get("password1")
        password2 = self.cleaned_data.get("password2")
        if password1 and password2 and password1 != password2:
            raise forms.ValidationError(
                self.error_messages['password_mismatch'],
                code='password_mismatch',
            )
        return password2

    def _post_clean(self):
        super()._post_clean()
        # Validate the password after self.instance is updated with form data
        # by super().
        password = self.cleaned_data.get('password2')
        if password:
            try:
                password_validation.validate_password(password, self.instance)
            except forms.ValidationError as error:
                self.add_error('password2', error)

    def save(self, commit=True):
        user = super().save(commit=False)
        user.set_password(self.cleaned_data["password1"])
        if commit:
            user.save()
        return user


class UserChangeForm(forms.ModelForm):
    password = ReadOnlyPasswordHashField(
        label=_("Password"),
        help_text=_(
            "Raw passwords are not stored, so there is no way to see this "
            "user's password, but you can change the password using "
            "<a href=\"{}\">this form</a>."
        ),
    )

    class Meta:
        model = User
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        password = self.fields.get('password')
        if password:
            password.help_text = password.help_text.format('../password/')
        user_permissions = self.fields.get('user_permissions')
        if user_permissions:
            user_permissions.queryset = user_permissions.queryset.select_related('content_type')

    def clean_password(self):
        # Regardless of what the user provides, return the initial value.
        # This is done here, rather than on the field, because the
        # field does not have access to the initial value
        return self.initial["password"]


class UserAdmin(BaseUserAdmin):

    add_form_template = 'admin/auth/user/add_form.html'
    change_user_password_template = None
    fieldsets = (
        (None, {'fields': ('email', 'password')}),
        (_('Personal info'), {'fields': ('first_name', 'last_name')}),
        (_('Permissions'), {'fields': ('is_active', 'is_staff', 'is_superuser', 'is_email_verified',
                                       'groups', 'user_permissions')}),
        (_('Important dates'), {'fields': ('last_login', 'date_joined')}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('email', 'password1', 'password2'),
        }),
    )

    add_form = UserCreationForm2
    form = UserChangeForm

    list_display = ('email', 'first_name', 'last_name', 'is_staff')
    list_filter = ('is_staff', 'is_superuser', 'is_active', 'is_email_verified', 'groups')
    search_fields = ('first_name', 'last_name', 'email')
    ordering = ('email',)
    filter_horizontal = ('groups', 'user_permissions',)


# Now register the new UserAdmin...
admin.site.register(User, UserAdmin)
admin.site.register(ImageAsset)
admin.site.register(RichTextAsset)
admin.site.register(TextAsset)
admin.site.register(Article)


class ArticleOrderInline(admin.TabularInline):
    model = ArticleOrder
    extra = 2


class ArticleCollectionOrderInline(admin.TabularInline):
    model = ArticleCollectionOrder
    fk_name = 'container_article_collection'
    extra = 2


class ArticleCollectionAdmin(admin.ModelAdmin):
    inlines = (ArticleOrderInline, ArticleCollectionOrderInline)


admin.site.register(ArticleCollection, ArticleCollectionAdmin)



from rest_framework import status
from rest_framework.response import Response

class Err:
    def __init__(self, error_code, status=status.HTTP_401_UNAUTHORIZED):
        self.error_code = error_code
        self.status = status

    def __str__(self):
        return self.error_code

    def __call__(self, message=None):
        data = {
            'error_code': self.error_code,
            'detail': message or '',
        }
        
        return Response(data, self.status)


ERR_USER_EXISTS = Err("ERR_USER_EXISTS")
ERR_USER_NOT_FOUND = Err("ERR_USER_EXISTS")
ERR_AUTH_FAIL = Err("ERR_AUTH_FAIL")
ERR_INVALID_PASSWORD = Err("ERR_VALIDATION_ERROR")

ERR_SUPPLIER_NOT_FOUND = Err("ERR_SUPPLIER_NOT_FOUND")
ERR_TARIFF_NOT_FOUND = Err("ERR_TARIFF_NOT_FOUND")



import logging

from django.core.mail.backends.base import BaseEmailBackend
from django.core.mail import EmailMultiAlternatives
from django.core.exceptions import ImproperlyConfigured


try:
    from google.appengine.api import mail as aeemail
    from google.appengine.ext import deferred
except ImportError as err:
    raise ImproperlyConfigured(('Failed to import App Engine libraries: %s'
                                % err))


class EmailBackend(BaseEmailBackend):

    can_defer = False

    def send_messages(self, email_messages):
        num_sent = 0
        for message in email_messages:
            if self._send(message):
                num_sent += 1
        return num_sent

    def _copy_message(self, message):
        """Create and return App Engine EmailMessage class from message."""
        gmsg = aeemail.EmailMessage(sender=message.from_email,
                                    to=message.to,
                                    subject=message.subject,
                                    body=message.body)
        if message.extra_headers.get('Reply-To', None):
            gmsg.reply_to = message.extra_headers['Reply-To']
        if message.bcc:
            gmsg.bcc = list(message.bcc)
        if message.attachments:
            gmsg.attachments = [(a[0], a[1]) for a in message.attachments]
        if isinstance(message, EmailMultiAlternatives):  # look for HTML
            for content, mimetype in message.alternatives:
                if mimetype == 'text/html':
                    gmsg.html = content
                    break
        return gmsg

    def _send(self, message):
        try:
            msg = self._copy_message(message)
        except (ValueError, aeemail.InvalidEmailError) as err2:
            logging.warning(err2)
            if not self.fail_silently:
                raise
            return False
        if self.can_defer:
            deferred.defer(self._send_deferred, msg)
            return True
        try:
            msg.send()
        except aeemail.Error:
            if not self.fail_silently:
                raise
            return False
        return True

    def _send_deferred(self, msg):
        msg.send()



from google.cloud import datastore


class Settings:
    """
    Get sensitive data setting from DataStore.

    key:String -> value:String
    key:String -> Exception

    Thanks to: Martin Omander @ Stackoverflow
    https://stackoverflow.com/a/35261091/1463812
    """

    client = datastore.Client()

    @staticmethod
    def get(name, default=None):

        query = Settings.client.query(kind='Settings')
        query.add_filter('name', '=', name)

        l = list(query.fetch())
        retval = None
        if len(l)>0:
            retval = list(query.fetch())[0]

        if not retval:
            if default is not None:
                return default
            Settings.set(name, 'Not Set')
            raise Exception(('Setting %s not found in the database. A placeholder ' +
                             'record has been created. Go to the Developers Console for your app ' +
                             'in App Engine, look up the Settings record with name=%s and enter ' +
                             'its value in that record\'s value field.') % (name, name))

        if str(retval['value']) == 'True' or str(retval['value']) == 'true':
            return True

        if str(retval['value']) == 'False' or str(retval['value']) == 'false':
            return False

        return retval['value']

    @staticmethod
    def set(name, value):

        key = Settings.client.key('Settings')
        entity = datastore.Entity(key=key)
        entity.update({'name': name, 'value': value})
        Settings.client.put(entity)



from django.conf import settings


def GlobalConfig(request):
    return {'GA_ACCOUNT_ID': settings.GA_ACCOUNT_ID,
            'ALLOW_ROBOTS': settings.ALLOW_ROBOTS,
            'SITE_URL': settings.SITE_URL,
            'HJID': settings.HJID,
            'HJSV': settings.HJSV,
            'DEBUG': ('false', 'true')[settings.DEBUG]
            }




from django.core.exceptions import ImproperlyConfigured

try:
    import default
except ImportError as err:
    raise ImproperlyConfigured(('Failed to import App Engine libraries: %s'
                                % err))


class EmailBackend(default.EmailBackend):
    can_defer = True



from django.urls import path
from django.conf.urls import url, include
from . import views
from django.contrib.admin.views.decorators import staff_member_required
from rest_framework import routers

router = routers.DefaultRouter()

router.register(r'articles', views.ArticleViewSet)
router.register(r'article_collections', views.ArticleCollectionViewSet)
router.register(r'imageassets', views.ImageAssetViewSet)
router.register(r'textassets', views.TextAssetViewSet)
router.register(r'richtextassets', views.RichTextAssetViewSet)

app_name = 'ripple'

urlpatterns = [

    # Account
    path('account/status', views.AccountStatusView.as_view()),
    path('account/log_in', views.LoginView.as_view(), name='log_in'),
    path('account/log_out', views.LogoutView.as_view()),
    path('account/sign_up', views.SignUpView.as_view()),
    path('account/<int:pk>/', views.UserUpdateView.as_view(), name="user_detail"),

    # Passwords
    path('account/password/reset', views.PasswordResetRequestView.as_view(),
         name='password_reset_request'),
    path('account/password/reset_confirm', views.PasswordResetConfirmView.as_view(),
         name='password_reset_confirm'),
    path('account/password/change', views.PasswordChangeView.as_view()),

    # Verification
    path('account/send_verify_email', views.SendEmailVerificationView.as_view()),
    path('account/verify_email',
         views.ConfirmEmailView.as_view(), name='verify_email'),

    # TODO red _ review these urls, are they needed?
    path('update_session_data', views.update_session_data),
    path('unpack_token', views.unpack_token),
    path('get_session_data', views.get_session_data),
    path('session', views.session),

    path('upload', views.simple_upload),
    path('assets', views.assets, name='assets'),
    path('assets_admin', staff_member_required(views.assets_admin)),

    path('', include(router.urls)),

]
